/**
  *  \file client/cscreen.cc
  *  \brief Control Screens, and Related Things
  *
  *  This contains the main classes for control screens.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/cscreen.h"
#include "client/chartusr.h"
#include "client/dialogs/consoledlg.h"
#include "client/dialogs/processmgr.h"
#include "client/history.h"
#include "client/screens.h"
#include "client/screens/minescreen.h"
#include "client/widgets/decaymsg.h"
#include "client/widgets/selswitch.h"
#include "client/wscanchart.h"
#include "gfx/palette.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "resmgr/resmgr.h"
#include "ui/prefix.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "client/screens/ionscreen.h"
#include "client/screens/ufoscreen.h"

/************************** WControlScreen::Skin ***************************/

/** \class WControlScreen::Skin
    \brief Skin for control screens */

class WControlScreen::Skin : public UISkin {
 public:
    Skin()
        : UISkin(DARK_COLOR_SET),
          background(requestPixmap("bg.cscreen"))
        { }
    void drawBackground(GfxContext& ctx, GfxRect r);

 private:
    Ptr<GfxPixmap> background;
};

void
WControlScreen::Skin::drawBackground(GfxContext& ctx, GfxRect r)
{
    drawSolidBar(ctx.getCanvas(), r, standard_colors[COLOR_GRAYSCALE+5]);
    if (background) {
        GfxPoint pt(0, 0);
        if (ctx.getCanvas().computeOffset(*ui_screen, pt)) {
            // Area the pixmap can fill, in screen (and pixmap) coordinates
            GfxRect pixArea(0, 0, background->getWidth(), background->getHeight());

            // Area we want to fill, in screen coordinates
            GfxRect fillArea(r);
            fillArea.moveBy(pt);

            // Area we will fill, in screen (and pixmap) coordinates
            fillArea.intersect(pixArea);

            // Draw it
            background->blit(ctx.getCanvas(),
                             fillArea.x - pt.x,
                             fillArea.y - pt.y,
                             fillArea);
        }
    }
}

/***************************** WControlScreen ****************************/

/*
  Layout:

    ,-----------------------------------.
    |              ,------------------. |   <-- borders are 5 pixels
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              `------------------' |  <-- tile_new_y  \ 3 lines for
    |              [                  ] |  ___ chart_y2    / scanner widget
    |                                   |
    |                                   |    8 lines free here for tiles
    |                                   |
    |                                   |
    `-----------------------------------'  ---

    |  320 pixels  |
                chart_x1
 */

/** Constructor.
    \param selection object selection we're displaying. Used for iteration.
    \param screen_id this screen's identifier.

    Checklist for derived constructors:
    - create all tiles and add them with addTile(). You should include a
      WControlScreenHeaderTile.
    - create starchart widget and hand it to addChart() */
WControlScreen::WControlScreen(GObjectSelection& selection, WScreenId::ScreenId screen_id)
    : screen_id(screen_id),
      selection(selection),
      conn_pos_changed(selection.sig_current_changed.add(this, &WControlScreen::onCurrentChanged)),
      scanner(),
      chart(0)
{
    setSkin(holder.add(new Skin()));
    setExtent(GfxRect(*ui_screen), true);
    setGeometry();
    setState(st_Modal, true);

    const int scanner_height = 2*font_heights[FONT_NORMAL];

    /* Note: for focus handling to work, make sure that none of the
       widgets added here claims it can receive focus (i.e. all must
       have st_Disabled set). This makes sure that focus is given to
       the first tile with st_Disabled clear. */
    add(scanner, GfxRect(chart_x1+2,                chart_y2 + 5,
                         extent.w - chart_x1 - 5-2, scanner_height));
    add(holder.add(new UIPrefixArg()), GfxRect(0, 0, 0, 0));
    add(holder.add(new WSelectionSwitcher()), GfxRect(0, 0, 0, 0));

    pushHistoryScreenId(WScreenId(screen_id, selection.getCurrentIndex()));
}

WControlScreen::~WControlScreen()
{ }

/** Compute geometry parameters. */
void
WControlScreen::setGeometry()
{
    tile_width = 315;           // FIXME: should probably be related to font size somehow
    tile_new_y = extent.h - 8*font_heights[FONT_NORMAL];
    tile_y     = 5;
    tile_x     = 5;
    chart_y2   = tile_new_y - 3*font_heights[FONT_NORMAL];
    chart_x1   = tile_width + tile_x + 5;
}

/** Selection-change handler. This is called when the user picks a
    different object on this object's type. */
void
WControlScreen::onCurrentChanged()
{
    pushHistoryScreenId(WScreenId(screen_id, selection.getCurrentIndex()));
    update();
}

/** Starchart double-click. Inject a virtual key. */
void
WControlScreen::onChartDblclick()
{
    /* Check with current modifiers. If none found, check again
       without shift, then without all modifiers.

       For regular events, we automatically discount shift
       (xref ui/window.cc:simplifyEvent). */
    uint32_t mods = getCurrentModifierKeys();
    if (!checkKeymap(vk_DoubleClickLocation + mods, 0)) {
        if (!checkKeymap(vk_DoubleClickLocation + (mods & ~ss_Shift), 0)) {
            checkKeymap(vk_DoubleClickLocation, 0);
        }
    }
}

/** Add starchart widget. Must be called once. */
void
WControlScreen::addChart(WScannerChartWidget& chart)
{
    this->chart = &chart;
    add(chart, GfxRect(chart_x1, 5, extent.w - chart_x1 - 5, chart_y2 - 5));
    scanner.setChart(chart);
    chart.sig_dblclick.add(this, &WControlScreen::onChartDblclick);
}

/** Add a tile. */
void
WControlScreen::addNewTile(WTile* t)
{
    /* Remember it */
    WTile& tile = holder.add(t);

    /* Add it */
    UIWidget::LayoutInfo info;
    tile.getLayoutInfo(info);

    if (tile_y + info.pref_size.y > extent.h - 5) {
        /* tile does not fit in this column. Would it fit when we
           remove the heading? */
        if (tile.getDisplayState() == WTile::sNormal) {
            tile.setDisplayState(WTile::sContentOnly);
            tile.getLayoutInfo(info);
            if (tile_y + info.pref_size.y <= extent.h - 5)
                goto success;
            tile.setDisplayState(WTile::sNormal);
            tile.getLayoutInfo(info);
        }

        /* start new column */
        tile_y = tile_new_y;
        tile_x += tile_width;

        /* does it fit there? */
        if (tile_y + info.pref_size.y > extent.h - 5) {
            /* nope. Remove heading if there, but we cannot do
               anything else so let it be. */
            if (tile.getDisplayState() == WTile::sNormal) {
                tile.setDisplayState(WTile::sContentOnly);
                tile.getLayoutInfo(info);
            }
        }
    }
 success:
    GfxRect tile_size(tile_x, tile_y, tile_width, info.pref_size.y);
    tile.setExtent(tile_size, true);
    tile_y += info.pref_size.y + 2;

    add(tile, tile_size);
}

/** Define keymap for this screen. */
void
WControlScreen::setKeymap(IntKeymapRef keymap)
{
    this->keymap = keymap;
}

void
WControlScreen::drawContent(GfxCanvas& can)
{
    // FIXME: optimize!
    GfxContext ctx;
    ctx.useColorScheme(getSkin());
    ctx.useCanvas(can);
    drawBackground(ctx, extent);

    drawFrameUp(can, extent);
    drawFrameDown(can, GfxRect(chart_x1-1, 5-1,        extent.w - chart_x1 - 3, chart_y2 - 5+2));
    drawFrameDown(can, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
    can.update(extent);
}

bool
WControlScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* Check user-defined keymaps */
        if (checkKeymap(e.key.code, e.key.arg)) {
            return true;
        }
        /* Check hardwired commands */
        switch(e.key.code) {
         case 'y':
            if (chart->isValid())
                chart->doKeyboardMode();
            return true;
         case 'q' + ss_Alt:
            doProcessManager(selection.getCurrentObject());
            return true;
        }
    }
    return UIFixedWidget::handleEvent(e, second_pass);
}

/** Browse through objects. */
void
WControlScreen::doBrowse(bool forward, bool marked)
{
    int n;
    if (forward) {
        n = selection.getObjectType()->findNextIndexWrap(selection.getCurrentIndex(), marked);
    } else {
        n = selection.getObjectType()->findPreviousIndexWrap(selection.getCurrentIndex(), marked);
    }

    if (selection.getObjectType()->isValidIndex(n)) {
        selection.setCurrentIndex(n);
    }
}

/** Get scanner position. */
GPoint
WControlScreen::getScannerPosition() const
{
    return chart->getScannerTarget();
}

/** Run this control screen. */
void
WControlScreen::runScreen()
{
    update();
    runModal(*this);
}

IntValue*
WControlScreen::getProperty(IntUserInterfaceProperty p)
{
    switch (p) {
     case iuiScreenNumber:
        return makeIntValue(screen_id);
     case iuiScreenRegistered:
        return makeBoolValue(true);
     case iuiChartX:
        if (chart->isValid())
            return makeIntValue(chart->getCenterPosition().x);
        else
            return 0;
     case iuiChartY:
        if (chart->isValid())
            return makeIntValue(chart->getCenterPosition().y);
        else
            return 0;
     case iuiScanX:
        if (chart->isValid())
            return makeIntValue(chart->getScannerTarget().x);
        else
            return 0;
     case iuiScanY:
        if (chart->isValid())
            return makeIntValue(chart->getScannerTarget().y);
        else
            return 0;
     case iuiSimFlag:
        return makeBoolValue(false);
     case iuiKeymap:
        return makeKeymapValue(keymap);
    }
    return 0;
}

void
WControlScreen::setProperty(IntUserInterfaceProperty p, IntValue* v)
{
    int32_t n;
    switch (p) {
     case iuiScanX:
        if (checkIntArg(n, v, 0, 10000) && chart->isValid()) {
            chart->setScannerTarget(GPoint(n, chart->getScannerTarget().y));
        }
        break;
     case iuiScanY:
        if (checkIntArg(n, v, 0, 10000) && chart->isValid()) {
            chart->setScannerTarget(GPoint(chart->getScannerTarget().x, n));
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

bool
WControlScreen::processPreIdle()
{
    if (selection.getObjectType()->isEmpty()) {
        // We have nothing to display, exit
        postGoToScreen(*this, 0, 0);
        return true;
    } else {
        // No problem
        return false;
    }
}

bool
WControlScreen::processEvent(std::auto_ptr<IntUserInterfaceEvent> event)
{
    switch (event->command) {
     case iucPopupConsole:
        doConsole();
        return false;

     case iucExitRace:
     case iucExitClient:
     case iucGoToScreen:
     case iucGoToChart:
        // FIXME: optimize for we-already-are-on-this-screen
        // FIXME: select current object
        IntUserInterfaceChildBinding::parent.postEvent(event);
        return true;

     case iucUseKeymap:
        return IntUserInterfaceDialogBinding::processEvent(event);

     case iucGoToMinefield:
        doMinefieldScreen();
        return true;

     case iucGoToUfo:
        doUfoScreen();
        return true;

     case iucGoToStorm:
        doIonStormScreen();
        return true;

     case iucOverlayMessage:
        showDecayingMessage(IntValue::toStringOf(event->args[0], false));
        return false;

     case iucChartSetView:
        // Nothing to do, command refuses internally
        return false;

     case iucExitDialog:
        return false;
    }
    return false;
}
