/**
  *  \file game/objnote.cc
  *  \brief Note Objects
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/objnote.h"

/** Constructor.
    \param id Identifier. 0=divider (see class description)
    \param name Name shown to user  */
GNoteObject::GNoteObject(int id, string_t name)
    : id(id), name(name)
{ }

/** Copy constructor. */
GNoteObject::GNoteObject(const GNoteObject& other)
    : GObject(), id(other.id), name(other.name)
{ }

GNoteObject::~GNoteObject()
{ }

string_t
GNoteObject::getName(Name /*which*/) const
{
    return name;
}

int
GNoteObject::getId() const
{
    return id;
}

int
GNoteObject::getOwner() const
{
    return 0;
}

/***************************** GNoteContainer ****************************/

/** Constructor. */
GNoteContainer::GNoteContainer()
    : notes()
{ }

GNoteContainer::~GNoteContainer()
{ }

/** Object access.
    \param index 1-based index */
GNoteObject&
GNoteContainer::getObjectByIndex(int index)
{
    ASSERT(index > 0);
    return *notes[index-1];
}

GUniverse*
GNoteContainer::getUniverseByIndex(int /*index*/)
{
    return 0;
}

bool
GNoteContainer::isValidIndex(int index) const
{
    return index > 0
        && index <= int(notes.size())
        && notes[index-1]->getId() != 0;
}

int
GNoteContainer::getNextIndex(int index) const
{
    if (index < int(notes.size())) {
        return index+1;
    } else {
        return 0;
    }
}

int
GNoteContainer::getPreviousIndex(int index) const
{
    if (index == 0) {
        return notes.size();
    } else {
        return index-1;
    }
}

/** Add a new note.
    \param id Id number (0=divider)
    \param name Name
    \return index, such that getObject(return) yields the newly-added note */
int
GNoteContainer::add(int id, string_t name)
{
    notes.push_back(new GNoteObject(id, name));
    sig_set_changed.raise(0);
    return notes.size();
}

/** Add a new note. The note will be copied from \c obj.
    \return index, such that getObject(return) yields the newly-added note */
int
GNoteContainer::addCopyOf(const GNoteObject& obj)
{
    notes.push_back(new GNoteObject(obj));
    sig_set_changed.raise(0);
    return notes.size();
}

/** Clear. Forgets all notes. */
void
GNoteContainer::clear()
{
    ptr_vector<GNoteObject>().swap(notes);
    sig_set_changed.raise(0);
}

/** Swap. Exchange this container's contents with another one.
    Care must be taken because notes are usually referred to by GObjectReference's,
    which refer to the respective GNoteContainer instance and an index into
    that instance. Those GObjectReference's will get invalid upon the swap. */
void
GNoteContainer::swap(GNoteContainer& other)
{
    notes.swap(other.notes);
    sig_set_changed.raise(0);
    other.sig_set_changed.raise(0);
}
