/**
  *  \file game/trn.h
  *  \brief Turn file access
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_TRN_H
#define PCC_V2_GAME_TRN_H

#include "game/struct.h"
#include "game/timestamp.h"
#include "cpluslib/buffer.h"

class Stream;
class GRegInfo;

/** TRN Command codes. The names are the same as in UN-TRN, the file
    format list, and some utilities inspired by the above.
    Those marked "<-" have been renamed, which does not mean the list
    would be consistent now.*/
enum {
    /* Ship commands */
    tcm_ShipFIRST              = 1,
    tcm_ShipChangeFc           = 1,        // sid, 3 bytes FC
    tcm_ShipChangeSpeed        = 2,        // sid, 1 word
    tcm_ShipChangeWaypoint     = 3,        // sid, 2 words
    tcm_ShipChangeMission      = 4,        // sid, 1 word
    tcm_ShipChangePrimaryEnemy = 5,        // sid, 1 word
    tcm_ShipTowShip            = 6,        // sid, 1 word
    tcm_ShipChangeName         = 7,        // sid, 20 bytes
    tcm_ShipBeamDownCargo      = 8,        // sid, 7 words NTDMCS+id
    tcm_ShipTransferCargo      = 9,        // sid, 7 words NTDMCS+id
    tcm_ShipIntercept          = 10,       // sid, 1 word
    tcm_ShipChangeNeutronium   = 11,       // sid, 1 word
    tcm_ShipChangeTritanium    = 12,       // sid, 1 word
    tcm_ShipChangeDuranium     = 13,       // sid, 1 word
    tcm_ShipChangeMolybdenum   = 14,       // sid, 1 word
    tcm_ShipChangeSupplies     = 15,       // sid, 1 word
    tcm_ShipChangeColonists    = 16,       // sid, 1 word
    tcm_ShipChangeTorpedoes    = 17,       // sid, 1 word
    tcm_ShipChangeMoney        = 18,       // sid, 1 word
    tcm_ShipLAST               = 18,

    /* Planet commands */
    tcm_PlanetFIRST            = 21,
    tcm_PlanetChangeFc         = 21,       // pid, 3 bytes
    tcm_PlanetChangeMines      = 22,       // pid, 1 word <-
    tcm_PlanetChangeFactories  = 23,       // pid, 1 word
    tcm_PlanetChangeDefense    = 24,       // pid, 1 word
    tcm_PlanetChangeNeutronium = 25,       // pid, 1 dword
    tcm_PlanetChangeTritanium  = 26,       // pid, 1 dword
    tcm_PlanetChangeDuranium   = 27,       // pid, 1 dword
    tcm_PlanetChangeMolybdenum = 28,       // pid, 1 dword
    tcm_PlanetChangeColonists  = 29,       // pid, 1 dword
    tcm_PlanetChangeSupplies   = 30,       // pid, 1 dword
    tcm_PlanetChangeMoney      = 31,       // pid, 1 dword
    tcm_PlanetColonistTax      = 32,       // pid, 1 word
    tcm_PlanetNativeTax        = 33,       // pid, 1 word
    tcm_PlanetLAST             = 33,       // BuildBase is special
    tcm_PlanetBuildBase        = 34,       // pid, NO DATA

    /* Starbase commands */
    tcm_BaseFIRST              = 40,
    tcm_BaseChangeDefense      = 40,       // bid, 1 word
    tcm_BaseUpgradeEngineTech  = 41,       // bid, 1 word
    tcm_BaseUpgradeHullTech    = 42,       // bid, 1 word <-
    tcm_BaseUpgradeWeaponTech  = 43,       // bid, 1 word <-
    tcm_BaseBuildEngines       = 44,       // bid, 9 words
    tcm_BaseBuildHulls         = 45,       // bid, 20 words
    tcm_BaseBuildWeapons       = 46,       // bid, 10 words
    tcm_BaseBuildLaunchers     = 47,       // bid, 10 words
    tcm_BaseBuildTorpedoes     = 48,       // bid, 10 words
    tcm_BaseBuildFighters      = 49,       // bid, 1 word
    tcm_BaseFixRecycleShipId   = 50,       // bid, 1 word <-
    tcm_BaseFixRecycleShip     = 51,       // bid, 1 word action
    tcm_BaseChangeMission      = 52,       // bid, 1 word
    tcm_BaseBuildShip          = 53,       // bid, 7 words
    tcm_BaseUpgradeTorpTech    = 54,       // bid, 1 word
    tcm_BaseLAST               = 54,

    /* Rest */
    tcm_SendMessage            = 60,       // len, from, to, text
    tcm_ChangePassword         = 61,       // zero, 10 bytes
    tcm_SendBack               = 62        // recv, type, size, data
};

//! Turn File Handling.
class GTurnfile {
 public:
    /** Feature Flags. */
    enum {
        trnf_Winplan = 1,       ///< File contains Winplan trailer.
        trnf_Taccom  = 2        ///< File contains Taccom-style attachments.
    };

    /** Manifest Constants. */
    enum {
        CURRENT_VERSION = 1,    ///< We know TRN sub-version 1 (Host 3.22).
        MAX_ATTACH      = 10    ///< Maximum number of attachments (Taccom).
    };

    /** Command Types. */
    enum CommandType {
        UndefinedCommand,       ///< Command not known to us.
        ShipCommand,            ///< Ship command. First word is Id.
        PlanetCommand,          ///< Planet command. First word is Id.
        BaseCommand,            ///< Base command. First word is Id.
        OtherCommand            ///< Other command. Password, Message, Sendback.
    };

    /* Constructor and Destructor */
    GTurnfile(int player, GTimestamp time);
    GTurnfile(Stream& stream, bool full_parse = true);
    ~GTurnfile();

    /* Header accessors */
    int                getPlayer() const;
    GTimestamp         getTimestamp() const;
    int                getNumCommands() const;
    int                getFeatures() const;
    int                getVersion() const;

    void               setTimestamp(const GTimestamp& time);
    void               setVersion(int n);
    void               setFeatures(int f);

    /* Trailer access */
    int                tryGetTurnNr() const;
    void               setTemplock(const uint32_t* data);
    void               setRegInfo(const GRegInfo& reg, int turn_nr);

    /* Header structure accessors */
    const TTurnWindowsTrailer& getWindowsTrailer() const;
    const TTurnDosTrailer& getDosTrailer() const;
    const TTurnHeader& getTurnHeader() const;
    const TTaccomTurnHeader& getTaccomHeader() const;

    /* Command accessors */
    int                getCommandCode(int index) const;
    int                getCommandLength(int index) const;
    int                getCommandId(int index) const;
    CommandType        getCommandType(int index) const;
    int32_t            getCommandPosition(int index) const;
    const char*        getCommandName(int index) const;
    const char*        getCommandData(int index) const;

    /* Command definition accessors */
    static CommandType getCommandCodeType(int code); // FIXME: these names suck
    static const char* getCommandCodeName(int code);
    static int         getCommandCodeRecordIndex(int code);

    /* Modificators */
    void               addCommand(int cmd, int id);
    void               addCommand(int cmd, int id, const char* data, std::size_t len);
    void               addData(const char* data, std::size_t len);
    void               deleteCommand(int index);

    /* Maketurn */
    void               makeShipCommands(int id, const char* old, const char* neu);
    void               makePlanetCommands(int id, const char* old, const char* neu);
    void               makeBaseCommands(int id, const char* old, const char* neu);

    /* Structure access */
    void               sortCommands();
    void               update();
    void               updateTrailer();
    int32_t            computeTurnChecksum() const;

    /* Taccom access */
    int                addFile(const char* file_data, std::size_t len, const string_t& name);
    int                addFile(const Buffer<char>& data, const string_t& name);
    void               deleteFile(int index);
    int                getNumFiles() const;
    int                getTaccomTurnPlace() const;

    /* Output */
    void               write(Stream& stream);

 private:
    struct CommandDef;
    static const CommandDef command_defs[];
    class CommandCompare;
    friend class CommandCompare;

    /* Turn file structure */
    TTurnHeader         trn_header;     ///< TRN header.
    TTaccomTurnHeader   taccom_header;  ///< Taccom header (TRN directory). Verbatim from turn file (offsets 1-based).
    TTurnDosTrailer     dos_trailer;    ///< DOS trailer.
    TTurnWindowsTrailer win_trailer;    ///< Windows trailer.
    Buffer<char>        data;           ///< Miscellaneous data. The TRN, usually ;)
    Buffer<int32>       offsets;        ///< Offsets of commands, pointing into data. zero-based. NOT the pointer array from the turn file!
    int                 version;        ///< TRN file sub-version (Winplan only).
    int                 features;       ///< TRN file features (bitfield trnf_Xxx).
    int                 trn_placement;  ///< Taccom: place TRN before Nth attachment.

    /* Internal stuff */
    bool dirty;                 ///< True if data is dirty. If false, data is a valid turn file.

    void init(Stream& str, bool full_parse);
    void checkRange(Stream& stream, int32_t offset, int32_t length);
    void parseTurnfile(Stream& stream, int32_t offset, int32_t length);
    void parseTurnfileHeader(Stream& stream, int32_t offset, int32_t length);

    void updateTurnfile(Buffer<char>& data, Buffer<int32_t>& offsets);
    void makeCommands(int id, int low, int up, const char* old, const char* neu);
};


/** Get player number. */
inline int
GTurnfile::getPlayer() const
{
    return trn_header.player;
}

/** Get turn timestamp. */
inline GTimestamp
GTurnfile::getTimestamp() const
{
    return GTimestamp(trn_header.timestamp);
}

/** Get number of commands stored in this turn.
    This is not necessary the number of commands the turn will have when written to disk,
    since there might be some deleted or invalid ones. */
inline int
GTurnfile::getNumCommands() const
{
    return offsets.getSize();
}

/** Get bitfield of features active in this turn file. */
inline int
GTurnfile::getFeatures() const
{
    return features;
}

/** Get sub-version of turn file. Only valid for Winplan turns.
    Returns the "xy" in "file format 3.5xy" (currently either 0 or 1). */
inline int
GTurnfile::getVersion() const
{
    return version;
}

/** Get Winplan (v3.5) trailer.
    This header is only set if the turn file actually is in Winplan format. */
inline const TTurnWindowsTrailer&
GTurnfile::getWindowsTrailer() const
{
    return win_trailer;
}

/** Get DOS (v3.0) Trailer. */
inline const TTurnDosTrailer&
GTurnfile::getDosTrailer() const
{
    return dos_trailer;
}

/** Get turn header. */
inline const TTurnHeader&
GTurnfile::getTurnHeader() const
{
    return trn_header;
}

/** Get raw TacCom header. Since that currently is only used by
    un-trn, I didn't bother to write accessors for everything. */
inline const TTaccomTurnHeader&
GTurnfile::getTaccomHeader() const
{
    return taccom_header;
}

/** Get position of turn data in TacCom container. */
inline int
GTurnfile::getTaccomTurnPlace() const
{
    return trn_placement;
}

#endif
