/**
  *  \file int/if/listif.cc
  *  \brief Interpreter Game Interface: Listboxes
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/listif.h"
#include "int/if/ifutil.h"
#include "int/args.h"
#include "cpluslib/smartptr.h"
#include "int/interror.h"
#include "cpluslib/misc.h"
#include "util/strlist.h"
#include "int/context.h"
#include "int/propenum.h"
#include "int/compound.h"
#include "int/exec.h"
#include "defs.h"               // HAVE_SDL
#ifdef HAVE_SDL
#include "ui/listbox.h"
#endif

namespace {
    enum {
        ListboxCommand
    };

    enum IntListboxCommand {
        ilcAddItem,
        ilcRun
    };

    /** Listbox data. */
    struct IntListbox {
        IntListbox()
            : title(), current(-1), width(320), height(-1), help(), list()
            { }

        string_t title;
        int32_t current;
        int32_t width;
        int32_t height;
        string_t help;

        StringList list;
    };

    /** Listbox data as an interpreter object. */
    class IntListboxValue : public IntContext {
     public:
        IntListboxValue(Ptr<IntListbox> data);

        virtual int32_t lookup(const IntNameQuery& name);
        virtual void set(int32_t index, IntValue* value);
        virtual IntValue* get(int32_t index);
        virtual bool next();
        virtual IntContext* clone();
        virtual GObject* getObject();
        virtual void enumProperties(IntPropertyAcceptor& acceptor);

        // IntValue:
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);

     private:
        Ptr<IntListbox> data;
    };

    /** Listbox command closure. */
    class IntListboxClosure : public IntCallableValue {
     public:
        IntListboxClosure(Ptr<IntListbox> data, IntListboxCommand cmd);

        void      call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
        bool      isProcedureCall();
        int32_t   getDimension(int32_t);
        IntContext* makeFirstContext();

        // IntValue:
        string_t    toString(bool readable);
        void        store(IntTagNode&, Stream&);
        IntValue*   clone();
     private:
        Ptr<IntListbox> data;
        IntListboxCommand cmd;
    };

    const TIntMapping property_mapping[] = {
        { "ADDITEM", ilcAddItem, ListboxCommand, thProcedure },
        { "RUN",     ilcRun,     ListboxCommand, thProcedure },
    };
}

/**************************** IntListboxValue ****************************/

inline
IntListboxValue::IntListboxValue(Ptr<IntListbox> data)
    : data(data)
{ }

int32_t
IntListboxValue::lookup(const IntNameQuery& name)
{
    return lookupName(name, property_mapping, countof(property_mapping));
}

void
IntListboxValue::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntListboxValue::get(int32_t index)
{
    return new IntListboxClosure(data, IntListboxCommand(property_mapping[index].index));
}

bool
IntListboxValue::next()
{
    return false;
}

IntContext*
IntListboxValue::clone()
{
    return new IntListboxValue(data);
}

GObject*
IntListboxValue::getObject()
{
    return 0;
}

void
IntListboxValue::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(property_mapping, countof(property_mapping));
}

// IntValue:
string_t
IntListboxValue::toString(bool /*readable*/)
{
    return "#<listbox>";
}

void
IntListboxValue::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/*************************** IntListboxClosure ***************************/

inline
IntListboxClosure::IntListboxClosure(Ptr<IntListbox> data, IntListboxCommand cmd)
    : data(data),
      cmd(cmd)
{ }

void
IntListboxClosure::call(IntExecutionContext& exc, IntDataSegment& args, bool want_result)
{
    IntArgBlock a(args, 0, args.size());
    switch (cmd) {
     case ilcRun:
     {
#ifdef HAVE_SDL
         /* @q Run (Listbox Command)
            Shows the list box and let the user select an item.
            If the user confirms the selection, the chosen item's %id is stored in {UI.Result}.
            If the user cancels, {UI.Result} is set to EMPTY.

            @see Listbox()
            @since PCC 1.1.1, PCC2 1.99.25 */
         int height = data->height < 3 && data->height != -1
             ? 3
             : data->height;
         a.checkArgumentCount(0);
         UIStandardListbox box(0, height, data->width, data->list);
         box.scrollToKey(data->current);
         if (box.doStandardListbox(data->title, data->help)) {
             IntIntIntValue iv(box.getKey());
             exc.setVariable("UI.RESULT", &iv);
         } else {
             exc.setVariable("UI.RESULT", 0);
         }
         break;
#else
         throw GError(GError::eGraph, GError::eGraph);
#endif
     }
     case ilcAddItem:
     {
         /* @q AddItem id:Int, text:Str (Listbox Command)
            Add an item to the list box.
            The item will be added at the end.
            The %text is displayed on the listbox.
            The %id will be used to select an item and report the user selection.

            @see Listbox()
            @since PCC 1.1.1, PCC2 1.99.25 */
         a.checkArgumentCount(2);
         int32_t id;
         string_t text;
         if (checkIntArg(id, a.getNext()) && checkStringArg(text, a.getNext())) {
             data->list.add(id, text);
         }
         break;
     }
    }
    if (want_result) {
        exc.pushNewValue(0);
    }
}

bool
IntListboxClosure::isProcedureCall()
{
    return true;
}

int32_t
IntListboxClosure::getDimension(int32_t)
{
    return 0;
}

IntContext*
IntListboxClosure::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntListboxClosure::toString(bool /*readable*/)
{
    return "#<listbox-command>";
}

void
IntListboxClosure::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntListboxClosure::clone()
{
    return new IntListboxClosure(*this);
}


/* @q Listbox():Any (Function)
   @noproto
   | With Listbox(title:Str, Optional current:Int, width:Int, height:Int, help:Str) Do
   |   AddItem id:Int, text:Str
   |   Run
   | EndWith

   This command sequence creates a standard list box.
   It consists of three parts:

   - the <tt>With Listbox()</tt> part creates a blank, still invisible list box template;
   - the <tt>{AddItem (Listbox Command)|AddItem}</tt> part adds items to the list box.
     You can use any number of these.
     Items are identified by their %id value which is an integer between 0 and 2^31-1
     (PCC 1.1.6 and below accept only values up to 32767);
   - the <tt>{Run (Listbox Command)|Run}</tt> part finally displays the list box
     and lets the user choose from it.
     You can repeat %Run as often as you wish.

   The parameters are as follows:
 
   - %title: a string that is displayed in the title bar of the list box.
     This is the only mandatory parameter for %Listbox();
   - %current: the %id value of the entry which will be selected by default
     when the list box opens. When there's no item with that identifier, the
     first one will be selected;
   - %width: the width of the list box in pixels. Must be between 200 and 1000,
     default is 320;
   - %height: the height of the list box in lines. Must be between 3 and 100,
     default is number of items in list box.
     Pass -1 here to choose that default;
   - %help: the help page associated with this list box. See {UI.Help}.
 
   The %Run command actually displays the list box and lets the user choose
   from it. It sets the {UI.Result} variable to the identifier (%id) of the
   item chosen by the user, or to EMPTY if she canceled.
 
   Example: this is a simplified version of the "Set Primary Enemy" command:
   |   Local i, UI.Result
   |   With Listbox("Primary Enemy", Enemy$, 260, 12, 10026) Do
   |     AddItem 0, "0 - none"
   |     For i:=1 To 11 Do AddItem i, Player(i).Race.Short
   |     Run
   |     SetEnemy UI.Result
   |   EndWith
 
   Note: scripts can not suspend while a <tt>With Listbox</tt> block is active.
   @since PCC 1.1.1, PCC2 1.99.25 */
IntValue*
IFListbox(IntArgBlock& args)
{
    args.checkArgumentCount(1, 5);

    // Parse args
    Ptr<IntListbox> data(new IntListbox());
    if (!checkStringArg(data->title, args.getNext())) {
        return 0;
    }
    checkIntArg(data->current, args.getNext());
    checkIntArg(data->width, args.getNext(), 200, 1000);
    checkIntArg(data->height, args.getNext(), -1, 100);
    checkStringArg(data->help, args.getNext());

    // Produce result
    return new IntListboxValue(data);
}
