/**
  *  \file console.h
  *  \brief Console Declarations
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CONSOLE_H
#define PCC_V2_CONSOLE_H

#include <string>
#include <ostream>
#include "cpluslib/sigslot.h"
#include "cpluslib/string.h"

/** Importance of messages. Sorted by severity, i.e. higher values are
    more important. */
enum MessageClass {
    LOG_DEBUG,
    LOG_INFO,
    LOG_WARN,
    LOG_ERROR
};

enum MessageFormat {
    ConsoleRightJustified = 0x8000,
    ConsoleRecallable     = 0x4000,
    ConsoleCentered       = 0x2000,
    ConsoleColorMask      = 0x001F           /* UIColor, 0 means default */
};

class Console : public std::ostream {
    enum { DEFAULT_SCROLLBACK_SIZE = 2000 };
 public:
    struct Message {
        MessageClass mc;
        string_t text;
        int format;
        Message(MessageClass mc, string_t text, int format)
            : mc(mc), text(text), format(format)
            { }
        Message()
            : mc(LOG_DEBUG), text(), format(0)
            { }
    };
 private:
    class Streambuf;
    friend class Streambuf;
    Streambuf* buf;

    int scrollback_length;
    Message* scrollback_buffer;
    int buf_head, buf_fill;
    MessageClass min_severity;

    void writeLine(MessageClass mc, string_t text, int format = 0);
    inline Message& getLineInternal(int n) const;
 public:
    Console();

    virtual ~Console();

    void write(MessageClass mc, string_t text, int format = 0);

    void setMinimumSeverity(MessageClass mc);

    /** Get length of scrollback buffer. */
    int getScrollbackSize() const
        { return scrollback_length; }

    void setScrollbackSize(int n);

    /** Get number of lines currently in scrollback buffer. */
    int getScrollbackFill() const
        { return buf_fill; }

    const Message& getLine(int n) const;

    void clear();

    /** Main Display Hook. Called when the console contents changes.
        \param arg1 the console which calls it
        \param arg2 number of lines that were added anew (in terms of
                    getLine(), lines arg2-1 .. 0). If zero, the console
                    was cleared (or something else that needs complete
                    redraw). */
    Signal2<const Console&, int> console_changed;
};

extern Console console;

#endif
